; stem version -->  http://www.schnada.de/progn/hax.html#doedel
; cf. Readme

;;; non R5RS:
;;;   SRFI-1  append! assoc/3 member/3 pair-for-each
;;;   listuls pair-map
;;;   I/O:    error-output-port format
;;;           error
;;;   syntuls def-renames with-renames
;;; main procedures (documented in place)
;;;   ipret-infocule
;;;   compile-infocule
;;;   [extend-ic-grammar and expand-xinfocule]
;;;   expand-db, expand-dp
;;; Note on generated names
;;;   The (REALLY-)COMPILE-sth procedures (but not EXTEND-IC-GRAMMAR and
;;; EXPAND-XINFOCULE) cope with infocule tags (possibly generated names)
;;; relying on the ``official'' comparison predicate.  I guess it would have
;;; been --certainly shocking, but nonetheless:-- quite acceptable to skip
;;; the hygienic habits, always `,coerce tags into symbols, and compare with
;;; EQ?  However, I haven't read an account of Clinger/Rees explicit-renaming
;;; macros so far, and only skimmed the relevant sources.  To wit, I'm in
;;; the dark, and will stay there for a while.
;;;   As is, extended infocule forms may lead to death and distruction
;;; or not depending on the context of evaluation (,loaded or prompted).

(define dt-html4s 
  "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">")
(define dt-html4l
  "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">")
(define dt-html4f
  "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Frameset//EN\" \"http://www.w3.org/TR/html4/frameset.dtd\">")

(define sticky 'sticky)
(define comment 'hush)
(define exec 'begin)

(define puncts '(
  area base br col hr img input link meta
  ;; cruft
  basefont isindex
  ))


(define brackets '(
  ;; blockish-brackets
  abbr acronym address
  blockquote body button
  caption cite code colgroup
  dd div dl dt fieldset form
  h1 h2 h3 h4 h5 h6 head html
  label legend li map noscript
  object ol optgroup option p pre
  samp script select style
  table tbody td textarea tfoot th thead title tr ul
  ;; cruft
  applet center dir frame frameset iframe menu noframes

  ;; flat-brackets -- unbe<em>lie</em>vable!
  a b bdo big del dfn em i ins kbd q
  small span strong sub sup tt var
  ;; cruft
  font s strike u
  ))


;;; interpreter and auxiliary procedures

;; ipret-infocule : infocule -> any
;; displays E as HTML code on the current output port, reporting errors to
;; the error output port.  White space:  IPRET-INFOCULE ends the output
;; of any sub-infocule with white space iff it's not in a sticky context.
(define (ipret-infocule e)
  (really-ipret-infocule e #t))
  

;; really-ipret-infocule : infocule boolean -> any
;; ends the HTML code corresponding to any sub-infocule E' <= E with wsp
;; iff BREAK? and E' is not in a sticky context
(define (really-ipret-infocule e break?)
  ;; accessors for infocules of the form (hush . contents) or ...
  (define tag first)
  (define body cdr)
  ;; ... (tag attributes . contents)
  (define attributes second)
  (define contents cddr)

  (cond ((procedure? e)
	 (e)
	 (if break? (newline)))
	 
	((not (pair? e))
	 (format #t "~A" e)
	 (if break?
	     ;; some pretty printing
	     (display (if (string? e) #\newline #\space))))

	((eq? (tag e) sticky)
	 (for-each (lambda (e)
		     (really-ipret-infocule e #f))
		   (body e)))
	 
	((eq? (tag e) comment)
	 (display "<!-- ")
	 ;; the internal spacing doesn't matter that much, but anyway ...
	 (pair-for-each
	  (lambda (e:es)
	    (really-ipret-infocule
	     (car e:es) (if (null? (cdr e:es)) break? #t)))
	  
	  (body e))
	 (display " -->")
	 (if break? (newline)))
	
	((memq (tag e) puncts)
	 (ipret-punct (tag e) (attributes e) (contents e) break?))

	((memq (tag e) brackets)
	 (ipret-bracket (tag e) (attributes e) (contents e) break?))

	(else
	 (let* ((rename (lambda (x) x))
		(maybe-ic (expand-xinfocule e rename eq?)))
	   (if maybe-ic
	       (really-ipret-infocule maybe-ic break?)
	       (format (error-output-port)
		       "really-ipret-infocule: ungrammatical input ~A~%"
		       e))))))


(define (ipret-punct tag attributes contents break?)
  (format #t "<~A" tag)
  (for-each (lambda (y)
	      (if (pair? y)
		  (format #t " ~A=\"~A\"" (first y) (second y))
		  (format #t " ~A" y)))
	    attributes)
  (display ">")
  (if break? (newline))
  (if (not (null? contents))
      (format (error-output-port)
	      "ipret-punct: skip non-empty CONTENTS of ~A~%" tag)))


(define (ipret-bracket tag attributes contents break?)
  (ipret-punct tag attributes '() #f)
  (pair-for-each
   (lambda (e:es)
     (really-ipret-infocule (car e:es)
			    (if (null? (cdr e:es)) break? #t)))
   
   contents)
  (format #t "</~A>" tag)
  (if break? (display " \n")))


;;; 
;;; compiler and auxiliary functions
;;;   The auxiliary compilers all return list(S<expression>) rather than
;;; S(BEGIN <expression>+) lest we get a horrible nest of blocks.  (Some
;;; return singleton lists -- uniformity is so nicely uniform.)

;; S<infocule> {symbol --> S<identifier>}
;; {S<identifier> $<identifier> --> boolean}     (?)
;;   --> S<expression>
(define (compile-infocule e rename %eq?)
  (cons (rename 'begin)
	(really-compile-infocule e rename %eq? #t)))

;; ... --> list(S<expression>)
(define (really-compile-infocule e rename %eq? break?)
  ;; accessors for comment, exec, and ``dynamic forms''
  (define tag first)
  (define body cdr)
  ;; ... (tag attributes . contents)
  (define attributes second)
  (define contents cddr)

  (def-renames rename
    (%q 'quote) (%qq 'quasiquote)
    (%car 'car) (%cdr 'cdr) (%null? 'null?) (%pair-for-each 'pair-for-each)
    (%newline 'newline) (%display 'display) (%format 'format)
    (%really-ipret-infocule 'really-ipret-infocule))

  (cond ((not (pair? e))
	 (if break?
	     `((,%format #t "~A~A"
			 (,%q ,e)
			 ,(if (string? e) #\newline #\space)))
	     `((,%format #t "~A" (,%q ,e)))))

	((%eq? (tag e) 'unquote)
	 `((,%really-ipret-infocule ,(second e) ,break?)))
	     
	((%eq? (tag e) 'unquote-splicing)
	 `((,%pair-for-each
	    (lambda (e:es)
	      (,%really-ipret-infocule
	       (,%car e:es) (if (,%null? (,%cdr e:es)) ,break? #t)))
	    ,(second e))))

	((%eq? (tag e) exec)
	 (if break?
	     `(,@(body e) (,%newline))
	     (body e)))

	((%eq? (tag e) sticky)
	 (apply append!
		;; list(list(S<expression>))
		(map (lambda (e)
		       ;; list(S<expression>)
		       (really-compile-infocule e rename %eq? #f))
		     (body e))))

	((%eq? (tag e) comment)
	 `((,%display "<!-- ")
	   ,@(apply append!
	      ;; list(list(S<expression>))
	      (pair-map
	       (lambda (e:es)
		 ;; list(S<expression>)
		 (really-compile-infocule
		  (car e:es) rename %eq? (if (null? (cdr e:es)) break? #t)))
	       (body e)))
	   (,%display " -->")
	   ,@(if break? `((,%newline)) '())))
	
	((member (tag e) puncts %eq?)
	 (compile-punct
	  (tag e) (attributes e) (contents e) rename %eq? break?))
	
	((member (tag e) brackets %eq?)
	 (compile-bracket
	  (tag e) (attributes e) (contents e) rename %eq? break?))
	
	(else
	 (let ((maybe-ic (expand-xinfocule e rename %eq?)))
	   (if maybe-ic
	       (really-compile-infocule maybe-ic rename %eq? break?)
	       (begin
		 (format (error-output-port)
			 "really-compile-infocule: ungrammatical input ~A~%"
			 e)
		 '()))))
	))
	  


;; S<identifier> S<attributes> list(S<infocule>) ... --> list(S<expression>)
;; I don't bother to compile `static' ATTRIBUTES. Everything will be done
;; at runtime. 
(define (compile-punct tag attributes contents rename %eq? break?)
  (if (not (null? contents))
      (format (error-output-port)
 	      "compile-punct: skip non-() CONTENTS of ~A~%" tag))
  (with-renames rename
    ((%q 'quote) (%qq 'quasiquote) (%ip 'ipret-punct))
    `((,%ip (,%q ,tag) (,%qq ,attributes) '() ,break?))))


;; ... --> list(S<expression>)
(define (compile-bracket tag attributes contents rename %eq? break?)
  (with-renames rename (begin display format)
    `(,@(compile-punct tag attributes '() rename %eq? #f)
      ,@(apply append!
	       ;; list(list(S<expression>))
	       (pair-map
		(lambda (e:es)
		  ;; list(S<expression>)
		  (really-compile-infocule
		   (car e:es) rename %eq? (if (null? (cdr e:es)) break? #t)))
	  contents))
      (,format #t "</~A>" ',tag)
      ,@(if break? `((,display " \n")) '())))) ; definitively spaced


;;; 
;;; We're entering the zone of shameful little secrets ...

;; tabulates keywords and (LAMBDA (E R C) ...) expanders
;; kxs : list(pair(symbol,{S<expression> --> S<expression>}))
(define kxs '())

;; extend the infocule grammar understood by COMPILE/IPRET-INFOCULE
;; the tag or keyword, possibly a generated name, is converted into the
;; corresponding symbol: `,name = symbol.
(define (extend-ic-grammar k x)
  (set! kxs (cons (cons `,k x) kxs)))

;; converts IC's tag name->symbol and compares with EQ? 
(define (expand-xinfocule ic rename %eq?)
  (let ((a (assoc `,(car ic) kxs eq?)))
    (if a ((cdr a) ic rename eq?) #f)))
;   (let ((a (assoc (car ic) kxs %eq?)))
;     (if a ((cdr a) ic rename %eq?) #f)))



;;; 
;;; element (and comment) emitters (sans doedel)

(define (emit-komment thunk)
  (display "<!-- ")
  (thunk)
  (display " -->"))

(define (emit-punct tag attributes)
  (ipret-punct tag attributes '() #f))

(define (emit-bracket tag attributes thunk)
  (emit-punct tag attributes)
  (thunk)
  (format #t "</~A>" tag))

;; expander for macro define-bracket
;;   (define-bracket name tag) -->
;; (begin
;;   (define (name* attributes thunk) <take this job and shove it>) 
;;   (define-syntax name
;;     (syntax-rules ()
;;       ((name attributes e0 ...)
;;        (name* `attributes (lambda () #f e0 ...))))))
;; Note that the attributes form of the macro NAME is implicitely backquoted:
;;   (define-bracket emphasise EM)
;;   (let ((lng it))
;;     (emphasise ((LANG ,lng))
;;       (display "miracolo"))
;; displays
;;   <EM LANG="it">miracolo</EM>
(define (expand-db exp rename %eq?)
  (with-renames rename
    ((%lambda 'lambda) (%begin 'begin) (%define 'define) (%quote 'quote)
     (%define-syntax 'define-syntax) (%syntax-rules 'syntax-rules)
     (%emit-punct 'emit-punct) (%emit-bracket 'emit-bracket))
    (let* ((name (second exp))
	   (name* (string->symbol (format #f "~A*" name))))
      `(,%begin
	(,%define (,name* attributes thunk)
		  (,%emit-bracket (,%quote ,(third exp)) attributes thunk))
	
	(,%define-syntax ,name
          (,%syntax-rules ()
	    ((,name attributes e0 ...)
	     (,name* `attributes (,%lambda () #f e0 ...)))))))))


;; expander for macro DEFINE-PUNCT
;;   <as DEFINE-BRACKET without thunk and body>
(define (expand-dp exp rename %eq?)
  (with-renames rename
    ((%lambda 'lambda) (%begin 'begin) (%define 'define) (%quote 'quote)
     (%define-syntax 'define-syntax) (%syntax-rules 'syntax-rules)
     (%emit-punct 'emit-punct) (%emit-bracket 'emit-bracket))
    (let* ((name (second exp))
	   (name* (string->symbol (format #f "~A*" name))))
      `(,%begin
	(,%define (,name* attributes)
		  (,%emit-punct (,%quote ,(third exp)) attributes))

	(,%define-syntax ,name
          (,%syntax-rules ()
	    ((,name attributes)
	     (,name* `attributes))))))))
