; stem version --> http://www.schnada.de/progn/hax.html#doedel
; cf. Readme

;;; syntactic utilities

;; doesn't support the mixed style (w/r ((v0 w0) v1 &c.) &c.)
(define-syntax with-renames
  (syntax-rules ()
    ((with-renames rename ((var sym) ...) e0 ...)
     (let ((var (rename sym)) ...) e0 ...))
    ((with-renames rename (n ...) e0 ...)
     (with-renames rename ((n 'n) ...) e0 ...))))

;; within expressions, always put DEFINEs above DEF-RENAMEs
(define-syntax def-renames
  (syntax-rules ()
    ((def-renames r (var sym) ...)
     (begin
       (define var (r sym)) ...))
    ((def-renames r n ...)
     (def-renames r (n 'n) ...))))



;;; list utilities

;; (list(t0) --> t1) list(t0) --> list(t1)
(define (pair-map f xs)
  (map! f (pair-fold-right  cons '() xs)))

#! Example
(pair-map (lambda (x:xs) (if (null? (cdr x:xs)) 'finis (car x:xs)))
	  '(1 2 3 6))
==> (1 2 3 finis)
!#


;;; i/o utilities

;; output : any* --> any
;; (sooa output) : outlet-designator any* --> any
;; ... outlet-designator as with FORMAT ... .
(define (screw-on-outport-adapter output)
  (lambda (out . args)
    (if (eq? #f out)
	(let ((op (make-string-output-port)))
	  (with-current-output-port op (apply output args))
	  (string-output-port-output op))
	(with-current-output-port
	 (cond ((eq? #t out) (current-output-port))
	       ((output-port? out) out)
	       (else (error "expect OUT : outlet-designator")))
	 (apply output args)))))

;; ...
(define-syntax with-current-opout-tourp
  (syntax-rules ()
    ((with-current-oport ?out ?e0 ?e1 ...)
     ((screw-on-outport-adapter (lambda () ?e0 ?e1 ...)) ?out))))



;;; string utilities

;; string string [re] --> string
;; If S0 = PRE o POST0, S1 = PRE o POST1,
;;   and PRE is a (maximal) submatch RE ~ S0, then
;;   (relative-tail s0 s1 re) ==> post1
;; Default RE = (rx (* any))
;; Examples:
;;   (relative-tail "/ober/land" "/ober/gau" (rx (: (* any) "/"))) ==> "gau"
(define (relative-tail s0 s1 . rest)
  (let* ((re (:optional rest (rx (* any))))
	 (m0 (regexp-search re s0)))
    (string-drop s1 (match:end m0 0))))

;; defensive version of RELATIVE-TAIL
;; string string [re] --> string
;; (relative-tail s0 s1 re)
;;   if the arguments satisfy the precondition of RELATIVE-TAIL, then
;;     ==> (relative-tail s0 s1 re)
;;   else <see below>
(define (relative-tailx s0 s1 . rest)
  (let* ((re (:optional rest (rx (* any))))
	 (m0 (regexp-search re s0)))
    (if m0
	(let* ((sm (regexp-substitute #f m0 0))
	       (m1 (regexp-search (rx ,sm) s1)))
	  (if m1 (regexp-substitute #f m1 'post) #f))
	#f)))
